/*
 * Copyright (C) 2017 aszlig
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * @defgroup stream_service_systemd stream_service_systemd
 * @{ @ingroup stream
 */

#ifndef STREAM_SERVICE_SYSTEMD_H_
#define STREAM_SERVICE_SYSTEMD_H_

/**
 * Create a service instance for systemd sockets.
 *
 * @param uri		URI with FD identifier, must start with "systemd://"
 * @param backlog	size of the backlog queue (ignored)
 * @return			stream_service instance, NULL on failure
 */
stream_service_t *stream_service_create_systemd(char *uri, int backlog);

#endif /** STREAM_SERVICE_SYSTEMD_H_ @}*/
